/*
 * AIEngine a deep packet inspector reverse engineering engine.
 *
 * Copyright (C) 2013  Luis Campo Giralte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com> 2013
 *
 */
#include "GPRSProtocol.h"
#include <iomanip> // setw

namespace aiengine {

void GPRSProtocol::process_create_pdp_context(Flow *flow) {

	SharedPointer<GPRSInfo> gprs_info = flow->gprs_info.lock();

	if (!gprs_info) {
		gprs_info = gprs_info_cache_->acquire().lock();
                if (gprs_info) {
			flow->gprs_info = gprs_info;
                }
	}

	if (gprs_info) {
		gprs_create_pdp_hdr *cpd = reinterpret_cast<gprs_create_pdp_hdr*>(gprs_header_->data);
		u_char *extensions = &cpd->data[0];
		uint8_t token = extensions[0];

//		std::cout << "Code(" << (int)cpd->code << ")presence(" << (int)cpd->presence << ")" << std::endl;

		if (cpd->presence == 0x02) {
			gprs_info->setIMSI(cpd->un.reg.imsi);
			extensions = &cpd->un.reg.hdr[0];
			token = extensions[0];
		}else {
			// And extension header
			if (cpd->presence == 0x01) {
				extensions = &cpd->data[0];
				token = extensions[0];
				gprs_info->setIMSI(cpd->un.ext.imsi);
			}
		}

		if (token == 0x03) { // Routing Area Identity Header
			gprs_create_pdp_hdr_routing *rhdr = reinterpret_cast<gprs_create_pdp_hdr_routing*>(extensions);
			extensions = &rhdr->data[0];
			token = extensions[0];
		}

		if (token == 0x0e) { // Recovery 
			extensions = &extensions[2];
			token = extensions[0];
		}
		if (token == 0x0f) { 
			gprs_create_pdp_hdr_ext *hext = reinterpret_cast<gprs_create_pdp_hdr_ext*>(&extensions[2]);
			extensions = &hext->data[0];
			token = extensions[0];

			if (token == 0x1a) { // Charging Characteristics
				token = extensions[3];
				extensions = &extensions[4];	
			} else {
				extensions = &extensions[1];	
			}
			if (token == 0x80) {
				uint16_t length = ntohs((extensions[1] << 8) + extensions[0]);
				if (length == 2) {
					uint8_t type_org = extensions[2];
					uint8_t type_num = extensions[3];
					// type_num eq 0x21 is IPv4
					// type_num eq 0x57 is IPv6
					
					gprs_info->setPdpTypeNumber(type_num);
				}
			}

		}
	}
}

void GPRSProtocol::processFlow(Flow *flow) {

	int bytes = flow->packet->getLength();
        total_bytes_ += bytes;
	++total_packets_;

        if (mux_.lock()&&(bytes > 0)) {
       
		uint8_t type = gprs_header_->type; 
		
		// just forward the packet if contains data
		if (type == T_PDU) {
			MultiplexerPtr mux = mux_.lock();

			Packet *packet = flow->packet;
			Packet gpacket;
			
			gpacket.setPayload(packet->getPayload());
			gpacket.setPrevHeaderSize(header_size);
			gpacket.setPayloadLength(packet->getLength());
			gpacket.setPacketTime(packet->getPacketTime());

			mux->setNextProtocolIdentifier(ETHERTYPE_IP); 
			mux->forwardPacket(gpacket);
		} else if (type == CREATE_PDP_CONTEXT_REQUEST) {
			process_create_pdp_context(flow);
		}
         }

}

void GPRSProtocol::statistics(std::basic_ostream<char>& out) {

	if (stats_level_ > 0) {
		out << "GPRSProtocol(" << this << ") statistics" << std::dec << std::endl;
		out << "\t" << "Total packets:          " << std::setw(10) << total_packets_ <<std::endl;
		out << "\t" << "Total bytes:        " << std::setw(14) << total_bytes_ <<std::endl;
		if (stats_level_ > 1) { 
			out << "\t" << "Total validated packets:" << std::setw(10) << total_validated_packets_ <<std::endl;
			out << "\t" << "Total malformed packets:" << std::setw(10) << total_malformed_packets_ <<std::endl;
			if (stats_level_ > 2) {
				if (mux_.lock())
					mux_.lock()->statistics(out);
				if (flow_forwarder_.lock())
					flow_forwarder_.lock()->statistics(out);

				if (stats_level_ > 3) {
					if (gprs_info_cache_)
                                        	gprs_info_cache_->statistics(out);
				}
			}
		}
	}
}

} // namespace aiengine 

